@@ -49,10 +49,10 @@ bool CommandManager::module_defined(StringView module_name) const
 void CommandManager::register_module(String module_name, String commands)
 {
     auto module = m_modules.find(module_name);
-    if (module != m_modules.end() and module->value.loaded)
+    if (module != m_modules.end() and module->value.state != Module::State::Registered)
         throw runtime_error{format("module already loaded: '{}'", module_name)};
 
-    m_modules[module_name] = { false, std::move(commands) };
+    m_modules[module_name] = { Module::State::Registered, std::move(commands) };
 }
 
 void CommandManager::load_module(StringView module_name, Context& context)
@@ -60,13 +60,23 @@ void CommandManager::load_module(StringView module_name, Context& context)
     auto module = m_modules.find(module_name);
     if (module == m_modules.end())
         throw runtime_error{format("no such module: '{}'", module_name)};
-    if (module->value.loaded)
-        return;
+    switch (module->value.state)
+    {
+        case Module::State::Loading:
+            throw runtime_error(format("module '{}' loaded recursively", module_name));
+        case Module::State::Loaded: return;
+        case Module::State::Registered: default: break;
+    }
 
-    module->value.loaded = true;
-    Context empty_context{Context::EmptyContextFlag{}};
-    execute(module->value.commands, empty_context);
+    {
+        module->value.state = Module::State::Loading;
+        auto restore_state = on_scope_end([&] { module->value.state = Module::State::Registered; });
+
+        Context empty_context{Context::EmptyContextFlag{}};
+        execute(module->value.commands, empty_context);
+    }
     module->value.commands.clear();
+    module->value.state = Module::State::Loaded;
 
     context.hooks().run_hook(Hook::ModuleLoaded, module_name, context);
 }
@@ -101,7 +111,7 @@ Reader& Reader::operator++()
     return *this;
 }
 
-void Reader::next_byte()
+Reader& Reader::next_byte()
 {
     kak_assert(pos < str.end());
     if (*pos++ == '\n')
@@ -109,6 +119,7 @@ void Reader::next_byte()
         ++line;
         line_start = pos;
     }
+    return *this;
 }
 
 namespace
@@ -226,16 +237,16 @@ void skip_blanks_and_comments(Reader& reader)
 {
     while (reader)
     {
-        const Codepoint c = *reader;
+        const Codepoint c = *reader.pos;
         if (is_horizontal_blank(c))
-            ++reader;
+            reader.next_byte();
         else if (c == '\\' and reader.pos + 1 != reader.str.end() and
                  *(reader.pos + 1) == '\n')
-            ++(++reader);
+            reader.next_byte().next_byte();
         else if (c == '#')
         {
             while (reader and *reader != '\n')
-                ++reader;
+                reader.next_byte();
         }
         else
             break;
@@ -299,41 +310,10 @@ Token parse_percent_token(Reader& reader, bool throw_on_unterminated)
     }
 }
 
-auto expand_option(const Option& opt, std::true_type)
-{
-    return opt.get_as_string(Quoting::Raw);
-}
-
-auto expand_option(const Option& opt, std::false_type)
-{
-    return opt.get_as_strings();
-}
-
-auto expand_register(StringView reg, const Context& context, std::true_type)
-{
-    return context.main_sel_register_value(reg).str();
-}
-
-auto expand_register(StringView reg, const Context& context, std::false_type)
-{
-    return RegisterManager::instance()[reg].get(context) | gather<Vector<String>>();
-}
-
-String expand_arobase(ConstArrayView<String> params, std::true_type)
-{
-    return join(params, ' ', false);
-}
-
-Vector<String> expand_arobase(ConstArrayView<String> params, std::false_type)
-{
-    return {params.begin(), params.end()};
-}
-
 template<bool single>
 std::conditional_t<single, String, Vector<String>>
 expand_token(const Token& token, const Context& context, const ShellContext& shell_context)
 {
-    using IsSingle = std::integral_constant<bool, single>;
     auto& content = token.content;
     switch (token.type)
     {
@@ -354,21 +334,40 @@ expand_token(const Token& token, const Context& context, const ShellContext& she
         return {str};
     }
     case Token::Type::RegisterExpand:
-        return expand_register(content, context, IsSingle{});
+        if constexpr (single)
+            return context.main_sel_register_value(content).str();
+        else
+            return RegisterManager::instance()[content].get(context) | gather<Vector<String>>();
     case Token::Type::OptionExpand:
-        return expand_option(context.options()[content], IsSingle{});
+    {
+        auto& opt = context.options()[content];
+        if constexpr (single)
+            return opt.get_as_string(Quoting::Raw);
+        else
+            return opt.get_as_strings();
+    }
     case Token::Type::ValExpand:
     {
         auto it = shell_context.env_vars.find(content);
         if (it != shell_context.env_vars.end())
             return {it->value};
-        return {ShellManager::instance().get_val(content, context, Quoting::Kakoune)};
+
+        auto val = ShellManager::instance().get_val(content, context);
+        if constexpr (single)
+            return join(val, false, ' ');
+        else
+            return val;
     }
     case Token::Type::ArgExpand:
     {
         auto& params = shell_context.params;
         if (content == '@')
-            return expand_arobase(params, IsSingle{});
+        {
+            if constexpr (single)
+                return join(params, ' ', false);
+            else
+                return Vector<String>{params.begin(), params.end()};
+        }
 
         const int arg = str_to_int(content)-1;
         if (arg < 0)
@@ -401,10 +400,10 @@ Optional<Token> CommandParser::read_token(bool throw_on_unterminated)
     const char* start = m_reader.pos;
     auto coord = m_reader.coord();
 
-    const Codepoint c = *m_reader;
+    const char c = *m_reader.pos;
     if (c == '"' or c == '\'')
     {
-        start = (++m_reader).pos;
+        start = m_reader.next_byte().pos;
         QuotedResult quoted = parse_quoted(m_reader, c);
         if (throw_on_unterminated and not quoted.terminated)
             throw parse_error{format("unterminated string {0}...{0}", c)};
@@ -417,9 +416,9 @@ Optional<Token> CommandParser::read_token(bool throw_on_unterminated)
         auto token = parse_percent_token(m_reader, throw_on_unterminated);
         return token;
     }
-    else if (is_command_separator(*m_reader))
+    else if (is_command_separator(c))
     {
-        ++m_reader;
+        m_reader.next_byte();
         return Token{Token::Type::CommandSeparator,
                      m_reader.pos - line.begin(), coord, {}};
     }
@@ -429,7 +428,7 @@ Optional<Token> CommandParser::read_token(bool throw_on_unterminated)
         {
             auto next = m_reader.peek_next();
             if (next == '%' or next == '\'' or next == '"')
-                ++m_reader;
+                m_reader.next_byte();
         }
         return Token{Token::Type::Raw, start - line.begin(),
                      coord, parse_unquoted(m_reader)};
@@ -488,13 +487,10 @@ struct command_not_found : runtime_error
         : runtime_error(format("no such command: '{}'", name)) {}
 };
 
-CommandManager::CommandMap::const_iterator
-CommandManager::find_command(const Context& context, StringView name) const
+StringView resolve_alias(const Context& context, StringView name)
 {
     auto alias = context.aliases()[name];
-    StringView cmd_name = alias.empty() ? name : alias;
-
-    return m_commands.find(cmd_name);
+    return alias.empty() ? name : alias;
 }
 
 void CommandManager::execute_single_command(CommandParameters params,
@@ -513,7 +509,7 @@ void CommandManager::execute_single_command(CommandParameters params,
     auto pop_cmd = on_scope_end([this] { --m_command_depth; });
 
     ParameterList param_view(params.begin()+1, params.end());
-    auto command_it = find_command(context, params[0]);
+    auto command_it = m_commands.find(resolve_alias(context, params[0]));
     if (command_it == m_commands.end())
         throw command_not_found(params[0]);
 
@@ -588,7 +584,7 @@ Optional<CommandInfo> CommandManager::command_info(const Context& context, Strin
          tokens.front().type != Token::Type::RawQuoted))
         return {};
 
-    auto cmd = find_command(context, tokens.front().content);
+    auto cmd = m_commands.find(resolve_alias(context, tokens.front().content));
     if (cmd == m_commands.end())
         return {};
 
@@ -641,7 +637,7 @@ Completions CommandManager::complete_command_name(const Context& context, String
 Completions CommandManager::complete_module_name(StringView query) const
 {
     return {0, query.length(),
-            Kakoune::complete(query, query.length(), m_modules | filter([](auto&& item) { return not item.value.loaded; })
+            Kakoune::complete(query, query.length(), m_modules | filter([](auto&& item) { return item.value.state == Module::State::Registered; })
                                                                | transform(&ModuleMap::Item::key))};
 }
 
@@ -730,7 +726,7 @@ Completions CommandManager::complete(const Context& context,
             flags |= CompletionFlags::Start;
         }
 
-        auto command_it = find_command(context, command_name);
+        auto command_it = m_commands.find(resolve_alias(context, command_name));
         if (command_it == m_commands.end())
             return Completions{};
 
@@ -755,7 +751,9 @@ Completions CommandManager::complete(const Context& context,
         if (not (completions.flags & Completions::Flags::Quoted) and token.type == Token::Type::Raw)
         {
             for (auto& c : completions.candidates)
-                c = (not c.empty() and contains("%'\"", c[0]) ? "\\" : "") + escape(c, "; \t", '\\');
+                c = (not c.empty() and c[0] == '%') or
+                        any_of(c, [](auto i) { return contains("; \t'\"", i); }) ?
+                            format("'{}'", replace(c, "'", "''")) : c;
         }
 
         return completions;
@@ -785,7 +783,7 @@ Completions CommandManager::complete(const Context& context,
             flags |= CompletionFlags::Start;
         }
 
-        auto command_it = find_command(context, command_name);
+        auto command_it = m_commands.find(resolve_alias(context, command_name));
         if (command_it != m_commands.end() and command_it->value.completer)
             return command_it->value.completer(
                 context, flags, params.subrange(1),
