@@ -92,35 +92,40 @@ struct MouseHandler
         Buffer& buffer = context.buffer();
         BufferCoord cursor;
         auto& selections = context.selections();
-        constexpr auto modifiers = Key::Modifiers::Control | Key::Modifiers::Alt | Key::Modifiers::Shift;
+        constexpr auto modifiers = Key::Modifiers::Control | Key::Modifiers::Alt | Key::Modifiers::Shift | Key::Modifiers::MouseButtonMask;
         switch ((key.modifiers & ~modifiers).value)
         {
-        case Key::Modifiers::MousePressRight:
-            m_dragging = false;
-            cursor = context.window().buffer_coord(key.coord());
-            if (key.modifiers & Key::Modifiers::Control)
-                selections = {{selections.begin()->anchor(), cursor}};
-            else
-                selections.main() = {selections.main().anchor(), cursor};
-            selections.sort_and_merge_overlapping();
-            return true;
-
-        case Key::Modifiers::MousePressLeft:
-            m_dragging = true;
-            m_anchor = context.window().buffer_coord(key.coord());
-            if (not (key.modifiers & Key::Modifiers::Control))
-                context.selections_write_only() = { buffer, m_anchor};
-            else
+        case Key::Modifiers::MousePress:
+            switch (key.mouse_button())
             {
-                size_t main = selections.size();
-                selections.push_back({m_anchor});
-                selections.set_main_index(main);
+            case Key::MouseButton::Right:
+                m_dragging = false;
+                cursor = context.window().buffer_coord(key.coord());
+                if (key.modifiers & Key::Modifiers::Control)
+                    selections = {{selections.begin()->anchor(), cursor}};
+                else
+                    selections.main() = {selections.main().anchor(), cursor};
                 selections.sort_and_merge_overlapping();
-            }
-            return true;
+                return true;
 
-        case Key::Modifiers::MouseReleaseLeft:
-        case Key::Modifiers::MouseReleaseRight:
+            case Key::MouseButton::Left:
+                m_dragging = true;
+                m_anchor = context.window().buffer_coord(key.coord());
+                if (not (key.modifiers & Key::Modifiers::Control))
+                    context.selections_write_only() = { buffer, m_anchor};
+                else
+                {
+                    size_t main = selections.size();
+                    selections.push_back({m_anchor});
+                    selections.set_main_index(main);
+                    selections.sort_and_merge_overlapping();
+                }
+                return true;
+
+            default: return true;
+            }
+
+        case Key::Modifiers::MouseRelease:
             if (not m_dragging)
                 return true;
             m_dragging = false;
@@ -138,7 +143,7 @@ struct MouseHandler
             return true;
 
         case Key::Modifiers::Scroll:
-            scroll_window(context, static_cast<int32_t>(key.key), not m_dragging);
+            scroll_window(context, static_cast<int32_t>(key.key), m_dragging);
             return true;
 
         default: return false;
@@ -273,7 +278,7 @@ public:
                         context.print_status(
                             { format("invalid register '{}'", *cp),
                               context.faces()["Error"] });
-                }, "enter target register", register_doc);
+                }, "enter target register", register_doc.str());
         }
         else
         {
@@ -817,7 +822,7 @@ public:
                     display();
                     m_line_changed = true;
                     m_refresh_completion_pending = true;
-                }, "enter register name", register_doc);
+                }, "enter register name", register_doc.str());
             display();
             return;
         }
@@ -1100,8 +1105,10 @@ private:
 class NextKey : public InputMode
 {
 public:
-    NextKey(InputHandler& input_handler, String name, KeymapMode keymap_mode, KeyCallback callback)
-        : InputMode(input_handler), m_name{std::move(name)}, m_callback(std::move(callback)), m_keymap_mode(keymap_mode) {}
+    NextKey(InputHandler& input_handler, String name, KeymapMode keymap_mode, KeyCallback callback,
+            Timer::Callback idle_callback)
+        : InputMode(input_handler), m_name{std::move(name)}, m_callback(std::move(callback)), m_keymap_mode(keymap_mode),
+          m_idle_timer{Clock::now() + get_idle_timeout(context()), std::move(idle_callback)} {}
 
     void on_key(Key key) override
     {
@@ -1122,9 +1129,10 @@ public:
     StringView name() const override { return m_name; }
 
 private:
-    String m_name;
-    KeyCallback m_callback;
-    KeymapMode m_keymap_mode;
+    String         m_name;
+    KeyCallback    m_callback;
+    KeymapMode     m_keymap_mode;
+    Timer          m_idle_timer;
 };
 
 class Insert : public InputMode
@@ -1274,7 +1282,7 @@ public:
                     if (not cp or key == Key::Escape)
                         return;
                     insert(RegisterManager::instance()[*cp].get(context()));
-                }, "enter register name", register_doc);
+                }, "enter register name", register_doc.str());
             update_completions = false;
         }
         else if (key == ctrl('n'))
@@ -1588,9 +1596,11 @@ void InputHandler::menu(Vector<DisplayLine> choices, MenuCallback callback)
     push_mode(new InputModes::Menu(*this, std::move(choices), std::move(callback)));
 }
 
-void InputHandler::on_next_key(StringView mode_name, KeymapMode keymap_mode, KeyCallback callback)
+void InputHandler::on_next_key(StringView mode_name, KeymapMode keymap_mode, KeyCallback callback,
+                               Timer::Callback idle_callback)
 {
-    push_mode(new InputModes::NextKey(*this, format("next-key[{}]", mode_name), keymap_mode, std::move(callback)));
+    push_mode(new InputModes::NextKey(*this, format("next-key[{}]", mode_name), keymap_mode, std::move(callback),
+                                      std::move(idle_callback)));
 }
 
 InputHandler::ScopedForceNormal::ScopedForceNormal(InputHandler& handler, NormalParams params)
@@ -1706,10 +1716,14 @@ std::pair<CursorMode, DisplayCoord> InputHandler::get_cursor_info() const
     return current_mode().get_cursor_info();
 }
 
+bool should_show_info(AutoInfo mask, const Context& context)
+{
+  return (context.options()["autoinfo"].get<AutoInfo>() & mask) and context.has_client();
+}
+
 bool show_auto_info_ifn(StringView title, StringView info, AutoInfo mask, const Context& context)
 {
-    if (not (context.options()["autoinfo"].get<AutoInfo>() & mask) or
-        not context.has_client())
+    if (not should_show_info(mask, context))
         return false;
 
     context.client().info_show(title.str(), info.str(), {}, InfoStyle::Prompt);
@@ -1722,31 +1736,34 @@ void hide_auto_info_ifn(const Context& context, bool hide)
         context.client().info_hide();
 }
 
-void scroll_window(Context& context, LineCount offset, bool adapt_cursor)
+void scroll_window(Context& context, LineCount offset, bool mouse_dragging)
 {
     Window& window = context.window();
     Buffer& buffer = context.buffer();
+    const LineCount line_count = buffer.line_count();
 
     DisplayCoord win_pos = window.position();
     DisplayCoord win_dim = window.dimensions();
 
+    if ((offset < 0 and win_pos.line == 0) or (offset > 0 and win_pos.line == line_count - 1))
+        return;
+
     const DisplayCoord max_offset{(win_dim.line - 1)/2, (win_dim.column - 1)/2};
     const DisplayCoord scrolloff =
         std::min(context.options()["scrolloff"].get<DisplayCoord>(), max_offset);
 
-    const LineCount line_count = buffer.line_count();
     win_pos.line = clamp(win_pos.line + offset, 0_line, line_count-1);
-    window.set_position(win_pos);
-
-    if (not adapt_cursor)
-        return;
 
     SelectionList& selections = context.selections();
-    const BufferCoord cursor = selections.main().cursor();
+    Selection& main_selection = selections.main();
+    const BufferCoord anchor = main_selection.anchor();
+    const BufferCoord cursor = main_selection.cursor();
 
-    auto line = clamp(cursor.line, win_pos.line + scrolloff.line,
+    auto cursor_off = mouse_dragging ? win_pos.line - window.position().line : 0;
+
+    auto line = clamp(cursor.line + cursor_off, win_pos.line + scrolloff.line,
                       win_pos.line + win_dim.line - 1 - scrolloff.line);
-    line = clamp(line, 0_line, line_count-1);
+    line = clamp(line, 0_line, buffer.line_count() - 1);
 
     using std::min; using std::max;
     // This is not exactly a clamp, and must be done in this order as
@@ -1754,7 +1771,13 @@ void scroll_window(Context& context, LineCount offset, bool adapt_cursor)
     auto col = min(max(cursor.column, buffer[line].byte_count_to(win_pos.column)),
                    buffer[line].length()-1);
 
-    selections = SelectionList{buffer, BufferCoord{line, col}};
+    BufferCoord new_cursor = { line, col };
+    BufferCoord new_anchor = (mouse_dragging or new_cursor == cursor) ? anchor : new_cursor;
+
+    window.set_position(win_pos);
+    main_selection = { new_anchor, new_cursor };
+
+    selections.sort_and_merge_overlapping();
 }
 
 }
