@@ -44,12 +44,21 @@ struct {
     unsigned int version;
     StringView notes;
 } constexpr version_notes[] = { {
+        20200901,
+        "» daemon mode does not fork anymore\n"
+    }, {
+        20200804,
+        "» {+u}User{} hook support\n"
+        "» Removed {+i}bold{} and {+i}italic{} faces from colorschemes\n"
+        "» Read from stdin support for clients\n"
+        "» {+u}rgba:RRGGBBAA{} faces and alpha blending\n"
+    }, {
         20200116,
-        "» {+u}InsertCompletionHide{} parameter is now the list of inserted ranges"
+        "» {+u}InsertCompletionHide{} parameter is now the list of inserted ranges\n"
     }, {
         20191210,
-        "» {+u}ModeChange{} parameter has changed to contain push/pop "
-        "{+ui}Mode{+u}Begin{}/{+ui}Mode{+u}End{} hooks were removed\n"
+        "» {+u}ModeChange{} parameter has changed to contain push/pop\n"
+        "» {+ui}Mode{+u}Begin{}/{+ui}Mode{+u}End{} hooks were removed\n"
     }, {
         20190701,
         "» {+u}%file\\{<filename>}{} expansions to read files\n"
@@ -107,7 +116,7 @@ void show_startup_info(Client* local_client, int last_version)
     DisplayLineList info;
     for (auto note : version_notes)
     {
-        if (note.version and note.version < last_version)
+        if (note.version and note.version <= last_version)
             continue;
 
         if (not note.version)
@@ -138,6 +147,9 @@ inline void write_stderr(StringView str) { try { write(STDERR_FILENO, str); } ca
 
 String runtime_directory()
 {
+    if (const char* runtime_directory = getenv("KAKOUNE_RUNTIME"))
+        return runtime_directory;
+
     char relpath[PATH_MAX+1];
     format_to(relpath, "{}../share/kak", split_path(get_kak_binary_path()).first);
     struct stat st;
@@ -156,179 +168,193 @@ String config_directory()
     return format("{}/.config/kak", homedir());
 }
 
+static auto main_sel_first(const SelectionList& selections)
+{
+    auto beg = &*selections.begin(), end = &*selections.end();
+    auto main = beg + selections.main_index();
+    using View = ConstArrayView<Selection>;
+    return concatenated(View{main, end}, View{beg, main});
+}
+
 static const EnvVarDesc builtin_env_vars[] = { {
         "bufname", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return context.buffer().display_name(); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {context.buffer().display_name()}; }
     }, {
         "buffile", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return context.buffer().name(); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {context.buffer().name()}; }
     }, {
         "buflist", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return join(BufferManager::instance() |
-                      transform(&Buffer::display_name) | transform(quoter(quoting)), ' ', false); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return BufferManager::instance() | transform(&Buffer::display_name) | gather<Vector>(); }
     }, {
         "buf_line_count", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string(context.buffer().line_count()); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(context.buffer().line_count())}; }
     }, {
         "timestamp", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string(context.buffer().timestamp()); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(context.buffer().timestamp())}; }
     }, {
         "history_id", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string((size_t)context.buffer().current_history_id()); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string((size_t)context.buffer().current_history_id())}; }
     }, {
         "selection", false,
-        [](StringView name, const Context& context, Quoting quoting)
+        [](StringView name, const Context& context) -> Vector<String>
         { const Selection& sel = context.selections().main();
-          return content(context.buffer(), sel); }
+          return {content(context.buffer(), sel)}; }
     }, {
         "selections", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return join(context.selections_content() | transform(quoter(quoting)), ' ', false); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return context.selections_content(); }
     }, {
         "runtime", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return runtime_directory(); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {runtime_directory()}; }
     }, {
         "config", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return config_directory(); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {config_directory()}; }
     }, {
         "version", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return version; }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {version}; }
     }, {
         "opt_", true,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return context.options()[name.substr(4_byte)].get_as_string(quoting); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return context.options()[name.substr(4_byte)].get_as_strings(); }
     }, {
         "main_reg_", true,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return context.main_sel_register_value(name.substr(9_byte)).str(); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {context.main_sel_register_value(name.substr(9_byte)).str()}; }
     }, {
         "reg_", true,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return join(RegisterManager::instance()[name.substr(4_byte)].get(context) |
-                      transform(quoter(quoting)), ' ', false); }
+        [](StringView name, const Context& context)
+        { return RegisterManager::instance()[name.substr(4_byte)].get(context) |
+                     gather<Vector<String>>(); }
     }, {
         "client_env_", true,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return context.client().get_env_var(name.substr(11_byte)).str(); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {context.client().get_env_var(name.substr(11_byte)).str()}; }
     }, {
         "session", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return Server::instance().session(); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {Server::instance().session()}; }
     }, {
         "client", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return context.name(); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {context.name()}; }
     }, {
         "client_pid", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string(context.client().pid()); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(context.client().pid())}; }
     }, {
         "client_list", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return join(ClientManager::instance() |
+        [](StringView name, const Context& context) -> Vector<String>
+        { return ClientManager::instance() |
                       transform([](const std::unique_ptr<Client>& c) -> const String&
-                               { return c->context().name(); }), ' ', false); }
+                                { return c->context().name(); }) | gather<Vector>(); }
     }, {
         "modified", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return context.buffer().is_modified() ? "true" : "false"; }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {context.buffer().is_modified() ? "true" : "false"}; }
     }, {
         "cursor_line", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string(context.selections().main().cursor().line + 1); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(context.selections().main().cursor().line + 1)}; }
     }, {
         "cursor_column", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string(context.selections().main().cursor().column + 1); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(context.selections().main().cursor().column + 1)}; }
     }, {
         "cursor_char_value", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
+        [](StringView name, const Context& context) -> Vector<String>
         { auto coord = context.selections().main().cursor();
           auto& buffer = context.buffer();
-          return to_string((size_t)utf8::codepoint(buffer.iterator_at(coord), buffer.end())); }
+          return {to_string((size_t)utf8::codepoint(buffer.iterator_at(coord), buffer.end()))}; }
     }, {
         "cursor_char_column", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
+        [](StringView name, const Context& context) -> Vector<String>
         { auto coord = context.selections().main().cursor();
-          return to_string(context.buffer()[coord.line].char_count_to(coord.column) + 1); }
+          return {to_string(context.buffer()[coord.line].char_count_to(coord.column) + 1)}; }
     }, {
         "cursor_display_column", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
+        [](StringView name, const Context& context) -> Vector<String>
         { auto coord = context.selections().main().cursor();
-          return to_string(get_column(context.buffer(),
-                                      context.options()["tabstop"].get<int>(),
-                                      coord) + 1); }
+          return {to_string(get_column(context.buffer(),
+                                       context.options()["tabstop"].get<int>(),
+                                       coord) + 1)}; }
     }, {
         "cursor_byte_offset", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
+        [](StringView name, const Context& context) -> Vector<String>
         { auto cursor = context.selections().main().cursor();
-          return to_string(context.buffer().distance({0,0}, cursor)); }
+          return {to_string(context.buffer().distance({0,0}, cursor))}; }
     }, {
         "selection_desc", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return selection_to_string(ColumnType::Byte, context.buffer(),
-                                     context.selections().main()); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {selection_to_string(ColumnType::Byte, context.buffer(), context.selections().main())}; }
     }, {
         "selections_desc", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return selection_list_to_string(ColumnType::Byte, context.selections()); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return main_sel_first(context.selections()) |
+                     transform([&buffer=context.buffer()](const Selection& sel) {
+                         return selection_to_string(ColumnType::Byte, buffer, sel);
+                     }) | gather<Vector>(); }
     }, {
         "selections_char_desc", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return selection_list_to_string(ColumnType::Codepoint, context.selections()); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return main_sel_first(context.selections()) |
+                     transform([&buffer=context.buffer()](const Selection& sel) {
+                         return selection_to_string(ColumnType::Codepoint, buffer, sel);
+                     }) | gather<Vector>(); }
     }, {
         "selections_display_column_desc", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return selection_list_to_string(ColumnType::DisplayColumn,
-                                          context.selections(),
-                                          context.options()["tabstop"].get<int>()); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return main_sel_first(context.selections()) |
+                     transform([&buffer=context.buffer(), tabstop=context.options()["tabstop"].get<int>()](const Selection& sel) {
+                         return selection_to_string(ColumnType::DisplayColumn, buffer, sel, tabstop);
+                     }) | gather<Vector>(); }
     }, {
         "selection_length", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string(char_length(context.buffer(), context.selections().main())); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(char_length(context.buffer(), context.selections().main()))}; }
     }, {
         "selections_length", false,
-        [](StringView name, const Context& context, Quoting quoting)
-        { return join(context.selections() |
-                      transform([&](const Selection& s)
-                               { return to_string(char_length(context.buffer(), s)); }), ' ', false); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return context.selections() |
+                     transform([&](const Selection& s) -> String {
+                         return to_string(char_length(context.buffer(), s));
+                     }) | gather<Vector<String>>(); }
     }, {
         "window_width", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string(context.window().dimensions().column); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(context.window().dimensions().column)}; }
     }, {
         "window_height", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return to_string(context.window().dimensions().line); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return {to_string(context.window().dimensions().line)}; }
     }, {
         "user_modes", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return join(context.keymaps().user_modes(), ' ', false); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return context.keymaps().user_modes(); }
     }, {
         "window_range", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
+        [](StringView name, const Context& context) -> Vector<String>
         {
             auto setup = context.window().compute_display_setup(context);
-            return format("{} {} {} {}", setup.window_pos.line, setup.window_pos.column,
-                                         setup.window_range.line, setup.window_range.column);
+            return {format("{} {} {} {}", setup.window_pos.line, setup.window_pos.column,
+                                          setup.window_range.line, setup.window_range.column)};
         }
     }, {
         "history", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return history_as_string(context.buffer().history(), quoting); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return history_as_strings(context.buffer().history()); }
     }, {
         "uncommitted_modifications", false,
-        [](StringView name, const Context& context, Quoting quoting) -> String
-        { return undo_group_as_string(context.buffer().current_undo_group(), quoting); }
+        [](StringView name, const Context& context) -> Vector<String>
+        { return undo_group_as_strings(context.buffer().current_undo_group()); }
     }
 };
 
@@ -336,25 +362,28 @@ void register_registers()
 {
     RegisterManager& register_manager = RegisterManager::instance();
 
-    for (auto c : StringView{"abcdefghijklmnopqrstuvwxyz\"^@"})
-        register_manager.add_register(c, std::make_unique<StaticRegister>());
+    for (Codepoint c : StringView{"abcdefghijklmnopqrstuvwxyz\"^@"})
+        register_manager.add_register(c, std::make_unique<StaticRegister>(String{c}));
 
-    for (auto c : StringView{"/|:\\"})
-        register_manager.add_register(c, std::make_unique<HistoryRegister>());
+    for (Codepoint c : StringView{"/|:\\"})
+        register_manager.add_register(c, std::make_unique<HistoryRegister>(String{c}));
 
     using StringList = Vector<String, MemoryDomain::Registers>;
 
     register_manager.add_register('%', make_dyn_reg(
+        "%",
         [](const Context& context)
         { return StringList{{context.buffer().display_name()}}; }));
 
     register_manager.add_register('.', make_dyn_reg(
+        ".",
         [](const Context& context) {
             auto content = context.selections_content();
             return StringList{content.begin(), content.end()};
          }));
 
     register_manager.add_register('#', make_dyn_reg(
+        "#",
         [](const Context& context) {
             const size_t count = context.selections().size();
             StringList res;
@@ -367,6 +396,7 @@ void register_registers()
     for (size_t i = 0; i < 10; ++i)
     {
         register_manager.add_register('0'+i, make_dyn_reg(
+            String{Codepoint('0'+i)},
             [i](const Context& context) {
                 StringList result;
                 for (auto& sel : context.selections())
@@ -647,9 +677,21 @@ int run_client(StringView session, StringView name, StringView client_init,
 {
     try
     {
+        Optional<int> stdin_fd;
+        if (not isatty(0))
+        {
+            // move stdin to another fd, and restore tty as stdin
+            stdin_fd = dup(0);
+            int tty = open("/dev/tty", O_RDONLY);
+            dup2(tty, 0);
+            close(tty);
+        }
+
         EventManager event_manager;
         RemoteClient client{session, name, make_ui(ui_type), getpid(), get_env_vars(),
-                            client_init, std::move(init_coord)};
+                            client_init, std::move(init_coord), stdin_fd};
+        stdin_fd.map(close);
+
         if (suspend)
             raise(SIGTSTP);
         while (not client.exit_status() and client.is_ui_ok())
@@ -687,21 +729,11 @@ int run_server(StringView session, StringView server_init,
                ConstArrayView<StringView> files)
 {
     static bool terminate = false;
-    if (flags & ServerFlags::Daemon)
+    set_signal_handler(SIGTERM, [](int) { terminate = true; });
+    if ((flags & ServerFlags::Daemon) and session.empty())
     {
-        if (session.empty())
-        {
-            write_stderr("-d needs a session name to be specified with -s\n");
-            return -1;
-        }
-        if (pid_t child = fork())
-        {
-            write_stderr(format("Kakoune forked to background, for session '{}'\n"
-                                "send SIGTERM to process {} for closing the session\n",
-                                session, child));
-            exit(0);
-        }
-        set_signal_handler(SIGTERM, [](int) { terminate = true; });
+        write_stderr("-d needs a session name to be specified with -s\n");
+        return -1;
     }
 
     EventManager        event_manager;
@@ -713,7 +745,7 @@ int run_server(StringView session, StringView server_init,
     CommandManager      command_manager;
     RegisterManager     register_manager;
     HighlighterRegistry highlighter_registry;
-    DefinedHighlighters defined_highlighters;
+    SharedHighlighters  defined_highlighters;
     ClientManager       client_manager;
     BufferManager       buffer_manager;
 
@@ -966,25 +998,19 @@ void signal_handler(int signal)
         case SIGSEGV: text = "SIGSEGV"; break;
         case SIGFPE:  text = "SIGFPE";  break;
         case SIGQUIT: text = "SIGQUIT"; break;
-        case SIGTERM: text = "SIGTERM"; break;
         case SIGPIPE: text = "SIGPIPE"; break;
     }
-    if (signal != SIGTERM)
-    {
-        auto msg = format("Received {}, exiting.\nPid: {}\nCallstack:\n{}",
-                          text, getpid(), Backtrace{}.desc());
-        write_stderr(msg);
-        notify_fatal_error(msg);
-    }
+    auto msg = format("Received {}, exiting.\nPid: {}\nCallstack:\n{}",
+                      text, getpid(), Backtrace{}.desc());
+    write_stderr(msg);
+    notify_fatal_error(msg);
 
     if (Server::has_instance())
         Server::instance().close_session();
     if (BufferManager::has_instance())
         BufferManager::instance().backup_modified_buffers();
 
-    if (signal == SIGTERM)
-        exit(-1);
-    else if (signal == SIGSEGV)
+    if (signal == SIGSEGV)
     {
         // generate core dump
         ::signal(SIGSEGV, SIG_DFL);
@@ -1150,8 +1176,14 @@ int main(int argc, char* argv[])
                 }
             }
             String new_files;
-            for (auto name : files)
-                new_files += format("edit '{}';", escape(real_path(name), "'", '\\'));
+            for (auto name : files) {
+                new_files += format("edit '{}'", escape(real_path(name), "'", '\\'));
+                if (init_coord) {
+                    new_files += format(" {} {}", init_coord->line + 1, init_coord->column + 1);
+                    init_coord.reset();
+                }
+                new_files += ";";
+            }
 
             return run_client(*server_session, {}, new_files + client_init, init_coord, ui_type, false);
         }
@@ -1160,10 +1192,12 @@ int main(int argc, char* argv[])
             StringView session = parser.get_switch("s").value_or(StringView{});
             try
             {
-                auto flags = (parser.get_switch("n") ? ServerFlags::IgnoreKakrc : ServerFlags::None) |
-                             (parser.get_switch("d") ? ServerFlags::Daemon : ServerFlags::None) |
-                             (parser.get_switch("ro") ? ServerFlags::ReadOnly : ServerFlags::None) |
-                             (argc == 1 and isatty(0) ? ServerFlags::StartupInfo : ServerFlags::None);
+                auto ignore_kakrc = (bool)parser.get_switch("n");
+                auto flags = (ignore_kakrc                                ? ServerFlags::IgnoreKakrc : ServerFlags::None) |
+                             (parser.get_switch("d")                      ? ServerFlags::Daemon      : ServerFlags::None) |
+                             (parser.get_switch("ro")                     ? ServerFlags::ReadOnly    : ServerFlags::None) |
+                             ((argc == 1 or (ignore_kakrc and argc == 2))
+                              and isatty(0)                               ? ServerFlags::StartupInfo : ServerFlags::None);
                 auto debug_flags = option_from_string(Meta::Type<DebugFlags>{}, parser.get_switch("debug").value_or(""));
                 return run_server(session, server_init, client_init, init_coord, flags, ui_type, debug_flags, files);
             }
