@@ -19,6 +19,7 @@
 #include <csignal>
 #include <sys/ioctl.h>
 #include <unistd.h>
+#include <strings.h>
 
 constexpr char control(char c) { return c & 037; }
 
@@ -252,7 +253,7 @@ int NCursesUI::Palette::get_color(Color color)
         return it->value;
     else if (m_change_colors and can_change_color() and COLORS > 16)
     {
-        kak_assert(color.color == Color::RGB);
+        kak_assert(color.isRGB());
         if (m_next_color > COLORS)
             m_next_color = 16;
         init_color(m_next_color,
@@ -264,7 +265,7 @@ int NCursesUI::Palette::get_color(Color color)
     }
     else
     {
-        kak_assert(color.color == Color::RGB);
+        kak_assert(color.isRGB());
         int lowestDist = INT_MAX;
         int closestCol = -1;
         for (int i = 0; i < std::min(256, COLORS); ++i)
@@ -346,7 +347,7 @@ NCursesUI::NCursesUI()
     start_color();
     use_default_colors();
 
-    set_raw_mode();
+    set_terminal_mode();
     enable_mouse(true);
 
     set_signal_handler(SIGWINCH, &signal_handler<&resize_pending>);
@@ -362,7 +363,7 @@ NCursesUI::~NCursesUI()
     enable_mouse(false);
     m_palette.set_change_colors(false);
     endwin();
-    tcsetattr(STDIN_FILENO, TCSAFLUSH, &m_original_termios);
+    restore_terminal_mode();
     set_signal_handler(SIGWINCH, SIG_DFL);
     set_signal_handler(SIGHUP, SIG_DFL);
     set_signal_handler(SIGTSTP, SIG_DFL);
@@ -381,7 +382,7 @@ void NCursesUI::suspend()
     sigemptyset(&unblock_sigtstp);
     sigaddset(&unblock_sigtstp, SIGTSTP);
     sigprocmask(SIG_UNBLOCK, &unblock_sigtstp, &old_mask);
-    tcsetattr(STDIN_FILENO, TCSAFLUSH, &m_original_termios);
+    restore_terminal_mode();
 
     raise(SIGTSTP); // suspend here
 
@@ -391,12 +392,12 @@ void NCursesUI::suspend()
 
     doupdate();
     check_resize(true);
-    set_raw_mode();
+    set_terminal_mode();
     m_palette.set_change_colors(change_color_enabled);
     enable_mouse(mouse_enabled);
 }
 
-void NCursesUI::set_raw_mode() const
+void NCursesUI::set_terminal_mode() const
 {
     termios attr = m_original_termios;
     attr.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON);
@@ -407,7 +408,19 @@ void NCursesUI::set_raw_mode() const
     attr.c_cflag |= CS8;
     attr.c_cc[VMIN] = attr.c_cc[VTIME] = 0;
 
-    tcsetattr(STDIN_FILENO, TCSAFLUSH, &attr);
+    tcsetattr(STDIN_FILENO, TCSANOW, &attr);
+    fputs("\033=", stdout);
+    // force enable report focus events
+    fputs("\033[?1004h", stdout);
+    fflush(stdout);
+}
+
+void NCursesUI::restore_terminal_mode() const
+{
+    tcsetattr(STDIN_FILENO, TCSAFLUSH, &m_original_termios);
+    fputs("\033>", stdout);
+    fputs("\033[?1004l", stdout);
+    fflush(stdout);
 }
 
 void NCursesUI::redraw(bool force)
@@ -593,22 +606,25 @@ Optional<Key> NCursesUI::get_next_key()
     if (not c)
         return {};
 
-    auto parse_key = [](unsigned char c) -> Key {
+    static constexpr auto convert = [](Codepoint c) -> Codepoint {
         if (c == control('m') or c == control('j'))
-            return {Key::Return};
+            return Key::Return;
         if (c == control('i'))
-            return {Key::Tab};
-        if (c == control('h'))
-            return {Key::Backspace};
+            return Key::Tab;
+        if (c == control('h') or c == 127)
+            return Key::Backspace;
+        return c;
+    };
+    auto parse_key = [](unsigned char c) -> Key {
+        if (Codepoint cp = convert(c); cp > 255)
+            return Key{cp};
         if (c == control('z'))
         {
             kill(0, SIGTSTP); // We suspend at this line
             return {};
         }
         if (c < 27)
-            return ctrl(Codepoint(c) - 1 + 'a');
-        if (c == 127)
-            return {Key::Backspace};
+            return ctrl(c - 1 + 'a');
 
        struct Sentinel{};
        struct CharIterator
@@ -654,19 +670,19 @@ Optional<Key> NCursesUI::get_next_key()
             return mod;
         };
 
-        auto mouse_button = [this](Key::Modifiers mod, Codepoint coord, bool left, bool release) {
-            auto mask = left ? 0x1 : 0x2;
+        auto mouse_button = [this](Key::Modifiers mod, Key::MouseButton button, Codepoint coord, bool release) {
+            auto mask = 1 << (int)button;
             if (not release)
             {
-                mod |= (m_mouse_state & mask) ? Key::Modifiers::MousePos : (left ? Key::Modifiers::MousePressLeft : Key::Modifiers::MousePressRight);
+                mod |= (m_mouse_state & mask) ? Key::Modifiers::MousePos : Key::Modifiers::MousePress;
                 m_mouse_state |= mask;
             }
             else
             {
-                mod |= left ? Key::Modifiers::MouseReleaseLeft : Key::Modifiers::MouseReleaseRight;
+                mod |= Key::Modifiers::MouseRelease;
                 m_mouse_state &= ~mask;
             }
-            return Key{mod, coord};
+            return Key{mod | Key::to_modifier(button), coord};
         };
 
         auto mouse_scroll = [this](Key::Modifiers mod, bool down) -> Key {
@@ -723,7 +739,7 @@ Optional<Key> NCursesUI::get_next_key()
             }
             return {};
         case 'u':
-            return masked_key(static_cast<Codepoint>(params[0]));
+            return masked_key(convert(static_cast<Codepoint>(params[0])));
         case 'Z': return shift(Key::Tab);
         case 'I': return {Key::FocusIn};
         case 'O': return {Key::FocusOut};
@@ -737,17 +753,15 @@ Optional<Key> NCursesUI::get_next_key()
             const int y = (sgr ? params[2] : next_char() - 32) - 1;
             auto coord = encode_coord({y - content_line_offset(), x});
             Key::Modifiers mod = parse_mask((b >> 2) & 0x7);
-            switch (b & 0x43)
+            switch (const int code = b & 0x43; code)
             {
-            case 0: return mouse_button(mod, coord, true, c == 'm');
-            case 2: return mouse_button(mod, coord, false, c == 'm');
+            case 0: case 1: case 2:
+                return mouse_button(mod, Key::MouseButton{code}, coord, c == 'm');
             case 3:
                 if (sgr)
                     return {};
-                if (m_mouse_state & 0x1)
-                    return mouse_button(mod, coord, true, true);
-                else if (m_mouse_state & 0x2)
-                    return mouse_button(mod, coord, false, true);
+                else if (int guess = ffs(m_mouse_state) - 1; 0 <= guess and guess < 3)
+                    return mouse_button(mod, Key::MouseButton{guess}, coord, true);
                 break;
             case 64: return mouse_scroll(mod, false);
             case 65: return mouse_scroll(mod, true);
@@ -1251,8 +1265,6 @@ void NCursesUI::enable_mouse(bool enabled)
     {
         // force SGR mode
         fputs("\033[?1006h", stdout);
-        // force enable report focus events
-        fputs("\033[?1004h", stdout);
         // enable mouse
         fputs("\033[?1000h", stdout);
         // force enable report mouse position
@@ -1262,7 +1274,6 @@ void NCursesUI::enable_mouse(bool enabled)
     {
         fputs("\033[?1002l", stdout);
         fputs("\033[?1000l", stdout);
-        fputs("\033[?1004l", stdout);
         fputs("\033[?1006l", stdout);
     }
     fflush(stdout);
