@@ -1,5 +1,7 @@
 #include "normal.hh"
 
+#include <functional>
+
 #include "buffer.hh"
 #include "buffer_manager.hh"
 #include "buffer_utils.hh"
@@ -90,14 +92,16 @@ void select(Context& context, T func)
     }
     else
     {
-        Vector<int> to_remove;
-        for (int i = 0; i < (int)selections.size(); ++i)
+        auto main_index = selections.main_index();
+        size_t new_size = 0;
+        for (size_t i = 0; i < selections.size(); ++i)
         {
             auto& sel = selections[i];
             auto res = func(context, sel);
             if (not res)
             {
-                to_remove.push_back(i);
+                if (i <= main_index and main_index != 0)
+                    --main_index;
                 continue;
             }
 
@@ -110,12 +114,15 @@ void select(Context& context, T func)
             }
             if (not res->captures().empty())
                 sel.captures() = std::move(res->captures());
+            if (i != new_size)
+                selections[new_size] = std::move(sel);
+            ++new_size;
         }
-
-        if (to_remove.size() == selections.size())
+        if (new_size == 0)
             throw no_selections_remaining{};
-        for (auto& i : to_remove | reverse())
-            selections.remove(i);
+
+        selections.set_main_index(main_index);
+        selections.remove_from(new_size);
     }
 
     selections.sort_and_merge_overlapping();
@@ -376,10 +383,10 @@ void view_commands(Context& context, NormalParams params)
             window.scroll(-std::max<ColumnCount>(1, count));
             break;
         case 'j':
-            window.scroll( std::max<LineCount>(1, count));
+            scroll_window(context,  std::max<LineCount>(1, count));
             break;
         case 'k':
-            window.scroll(-std::max<LineCount>(1, count));
+            scroll_window(context, -std::max<LineCount>(1, count));
             break;
         case 'l':
             window.scroll( std::max<ColumnCount>(1, count));
@@ -515,16 +522,18 @@ BufferCoord apply_diff(Buffer& buffer, BufferCoord pos, StringView before, Strin
 
     for_each_diff(lines_before.begin(), (int)lines_before.size(),
                   lines_after.begin(), (int)lines_after.size(),
-                  [&, posA = 0](DiffOp op, int len, int posB) mutable {
+                  [&, posA = 0, posB = 0](DiffOp op, int len) mutable {
         switch (op)
         {
         case DiffOp::Keep:
             pos = buffer.advance(pos, byte_count(lines_before, posA, len));
             posA += len;
+            posB += len;
             break;
         case DiffOp::Add:
             pos = buffer.insert(pos, {lines_after[posB].begin(),
                                       lines_after[posB + len - 1].end()}).end;
+            posB += len;
             break;
         case DiffOp::Remove:
             pos = buffer.erase(pos, buffer.advance(pos, byte_count(lines_before, posA, len)));
@@ -1774,11 +1783,17 @@ SelectionList read_selections_from_register(char reg, Context& context)
     if (content.size() < 2)
         throw runtime_error(format("register '{}' does not contain a selections desc", reg));
 
+    // Use the last two values for timestamp and main_index to allow the buffer
+    // name to have @ symbols
     struct error : runtime_error { error(size_t) : runtime_error{"expected <buffer>@<timestamp>@main_index"} {} };
-    const auto desc = content[0] | split<StringView>('@') | static_gather<error, 3>();
-    Buffer& buffer = BufferManager::instance().get_buffer(desc[0]);
-    const size_t timestamp = str_to_int(desc[1]);
-    size_t main = str_to_int(desc[2]);
+    auto end_content = content[0] | reverse() | split('@') | transform([] (auto bounds) {
+        return StringView{bounds.second.base(), bounds.first.base()};
+    }) | static_gather<error, 2, false>();
+
+    const size_t main = str_to_int(end_content[0]);
+    const size_t timestamp = str_to_int(end_content[1]);
+    const auto buffer_name = StringView{ content[0].begin (), end_content[1].begin () - 1 };
+    Buffer& buffer = BufferManager::instance().get_buffer(buffer_name);
 
     return selection_list_from_strings(buffer, ColumnType::Byte, content | skip(1), timestamp, main);
 }
@@ -1873,7 +1888,7 @@ void combine_selections(Context& context, SelectionList list, Func func, StringV
                                      list.set_main_index(sels.main_index());
                                  }
                                  func(context, std::move(list));
-                             }, title,
+                             }, title.str(),
                              "'a': append lists\n"
                              "'u': union\n"
                              "'i': intersection\n"
