@@ -309,7 +309,7 @@ find_opening(Iterator pos, const Container& container,
     {
         if (nestable)
         {
-            for (auto m [[gnu::unused]] : RegexIt{match[0].second, pos, container.begin(), container.end(), closing})
+            for (auto m [[maybe_unused]] : RegexIt{match[0].second, pos, container.begin(), container.end(), closing})
                 ++level;
         }
 
@@ -327,17 +327,12 @@ find_closing(Iterator pos, const Container& container,
              const Regex& opening, const Regex& closing,
              int level, bool nestable)
 {
-    MatchResults<Iterator> res;
-    if (regex_search(pos, container.end(), container.begin(), container.end(),
-                     res, opening) and res[0].first == pos)
-        pos = res[0].second;
-
     using RegexIt = RegexIterator<Iterator, RegexMode::Forward>;
     for (auto match : RegexIt{pos, container.end(), container.begin(), container.end(), closing})
     {
         if (nestable)
         {
-            for (auto m [[gnu::unused]] : RegexIt{pos, match[0].first, container.begin(), container.end(), opening})
+            for (auto m [[maybe_unused]] : RegexIt{pos, match[0].first, container.begin(), container.end(), opening})
                 ++level;
         }
 
@@ -355,6 +350,7 @@ find_surrounding(const Container& container, Iterator pos,
                  const Regex& opening, const Regex& closing,
                  ObjectFlags flags, int level)
 {
+    auto empty = [](const std::pair<Iterator, Iterator>& m) { return m.first == m.second; };
     const bool nestable = opening != closing;
     auto first = pos;
     auto last = pos;
@@ -365,18 +361,23 @@ find_surrounding(const Container& container, Iterator pos,
             first = (flags & ObjectFlags::Inner) ? res->second : res->first;
             if (flags & ObjectFlags::ToEnd) // ensure we find the matching end
             {
-                last = res->first;
+                last = empty(*res) ? std::next(res->second) : res->second;
                 level = 0;
             }
         }
         else
             return {};
     }
+    else if (MatchResults<Iterator> res;
+             regex_search(pos, container.end(), container.begin(), container.end(), res, opening) and
+             res[0].first == pos) // Skip opening match if pos lies on it
+        last = empty(res[0]) ? std::next(res[0].second) : res[0].second;
+
     if (flags & ObjectFlags::ToEnd)
     {
         if (auto res = find_closing(last, container, opening, closing, level, nestable))
-            last = (flags & ObjectFlags::Inner) ? utf8::previous(res->first, container.begin())
-                                                : utf8::previous(res->second, container.begin());
+            last = utf8::previous((flags & ObjectFlags::Inner) ? res->first : res->second,
+                                  container.begin());
         else
             return {};
     }
@@ -493,58 +494,63 @@ select_sentence(const Context& context, const Selection& selection,
 
     auto& buffer = context.buffer();
     BufferIterator first = buffer.iterator_at(selection.cursor());
+    BufferIterator last;
 
-    if (not (flags & ObjectFlags::ToEnd) and first != buffer.begin())
+    for (++count; count > 0; --count)
     {
-        BufferIterator prev_non_blank = first-1;
-        skip_while_reverse(prev_non_blank, buffer.begin(),
-                           [](char c) { return is_horizontal_blank(c) or is_eol(c); });
-        if (is_end_of_sentence(*prev_non_blank))
-            first = prev_non_blank;
-    }
-
-    BufferIterator last = first;
-
-    if (flags & ObjectFlags::ToBegin)
-    {
-        bool saw_non_blank = false;
-        while (first != buffer.begin())
+        if (not (flags & ObjectFlags::ToEnd) and first != buffer.begin())
         {
-            char cur = *first;
-            char prev = *(first-1);
-            if (not is_horizontal_blank(cur))
-                saw_non_blank = true;
-            if (is_eol(prev) and is_eol(cur))
+            BufferIterator prev_non_blank = first-1;
+            skip_while_reverse(prev_non_blank, buffer.begin(),
+                               [](char c) { return is_horizontal_blank(c) or is_eol(c); });
+            if (is_end_of_sentence(*prev_non_blank))
+                first = prev_non_blank;
+        }
+
+        if (last == BufferIterator{})
+            last = first;
+
+        if (flags & ObjectFlags::ToBegin)
+        {
+            bool saw_non_blank = false;
+            while (first != buffer.begin())
             {
-                ++first;
-                break;
-            }
-            else if (is_end_of_sentence(prev))
-            {
-                if (saw_non_blank)
+                char cur = *first;
+                char prev = *(first-1);
+                if (not is_horizontal_blank(cur))
+                    saw_non_blank = true;
+                if (is_eol(prev) and is_eol(cur) and first + 1 != buffer.end())
+                {
+                    ++first;
                     break;
-                else if (flags & ObjectFlags::ToEnd)
-                    last = first-1;
+                }
+                else if (is_end_of_sentence(prev))
+                {
+                    if (saw_non_blank)
+                        break;
+                    else if (flags & ObjectFlags::ToEnd)
+                        last = first-1;
+                }
+                --first;
             }
-            --first;
+            skip_while(first, buffer.end(), is_horizontal_blank);
         }
-        skip_while(first, buffer.end(), is_horizontal_blank);
-    }
-    if (flags & ObjectFlags::ToEnd)
-    {
-        while (last != buffer.end())
+        if (flags & ObjectFlags::ToEnd)
         {
-            char cur = *last;
-            if (is_end_of_sentence(cur) or
-                (is_eol(cur) and (last+1 == buffer.end() or is_eol(*(last+1)))))
-                break;
-            ++last;
-        }
-        if (not (flags & ObjectFlags::Inner) and last != buffer.end())
-        {
-            ++last;
-            skip_while(last, buffer.end(), is_horizontal_blank);
-            --last;
+            while (last != buffer.end())
+            {
+                char cur = *last;
+                if (is_end_of_sentence(cur) or
+                    (is_eol(cur) and (last+1 == buffer.end() or is_eol(*(last+1)))))
+                    break;
+                ++last;
+            }
+            if (not (flags & ObjectFlags::Inner) and last != buffer.end())
+            {
+                ++last;
+                skip_while(last, buffer.end(), is_horizontal_blank);
+                --last;
+            }
         }
     }
     return (flags & ObjectFlags::ToEnd) ? Selection{first.coord(), last.coord()}
@@ -557,51 +563,55 @@ select_paragraph(const Context& context, const Selection& selection,
 {
     auto& buffer = context.buffer();
     BufferIterator first = buffer.iterator_at(selection.cursor());
+    BufferIterator last;
 
-    if (not (flags & ObjectFlags::ToEnd) and first.coord() > BufferCoord{0,1} and
-        *(first-1) == '\n' and *(first-2) == '\n')
-        --first;
-    else if ((flags & ObjectFlags::ToEnd) and
-             first != buffer.begin() and (first+1) != buffer.end() and
-             *(first-1) == '\n' and *first == '\n')
-        ++first;
-
-    BufferIterator last = first;
-
-    if ((flags & ObjectFlags::ToBegin) and first != buffer.begin())
+    for (++count; count > 0; --count)
     {
-        skip_while_reverse(first, buffer.begin(),
-                           [](Codepoint c){ return is_eol(c); });
-        if (flags & ObjectFlags::ToEnd)
-            last = first;
-        while (first != buffer.begin())
-        {
-            char cur = *first;
-            char prev = *(first-1);
-            if (is_eol(prev) and is_eol(cur))
-            {
-                ++first;
-                break;
-            }
+        if (not (flags & ObjectFlags::ToEnd) and first.coord() > BufferCoord{0,1} and
+            is_eol(*(first-1)) and first-1 != buffer.begin() and is_eol(*(first-2)))
             --first;
-        }
-    }
-    if (flags & ObjectFlags::ToEnd)
-    {
-        if (last != buffer.end() and is_eol(*last))
-            ++last;
-        while (last != buffer.end())
+        else if ((flags & ObjectFlags::ToEnd) and
+                 first != buffer.begin() and (first+1) != buffer.end() and
+                 is_eol(*(first-1)) and is_eol(*first))
+            ++first;
+        if (last == BufferIterator{})
+            last = first;
+
+        if ((flags & ObjectFlags::ToBegin) and first != buffer.begin())
         {
-            if (last != buffer.begin() and is_eol(*last) and is_eol(*(last-1)))
-            {
-                if (not (flags & ObjectFlags::Inner))
-                    skip_while(last, buffer.end(),
+            skip_while_reverse(first, buffer.begin(),
                                [](Codepoint c){ return is_eol(c); });
-                break;
+            if (flags & ObjectFlags::ToEnd)
+                last = first;
+            while (first != buffer.begin())
+            {
+                char cur = *first;
+                char prev = *(first-1);
+                if (is_eol(prev) and is_eol(cur))
+                {
+                    ++first;
+                    break;
+                }
+                --first;
             }
-            ++last;
         }
-        --last;
+        if (flags & ObjectFlags::ToEnd)
+        {
+            if (last != buffer.end() and is_eol(*last))
+                ++last;
+            while (last != buffer.end())
+            {
+                if (last != buffer.begin() and is_eol(*last) and is_eol(*(last-1)))
+                {
+                    if (not (flags & ObjectFlags::Inner))
+                        skip_while(last, buffer.end(),
+                                   [](Codepoint c){ return is_eol(c); });
+                    break;
+                }
+                ++last;
+            }
+            --last;
+        }
     }
     return (flags & ObjectFlags::ToEnd) ? Selection{first.coord(), last.coord()}
                                         : Selection{last.coord(), first.coord()};
